from snowflake.snowpark import DataFrame, Row, DataFrameReader
from snowflake.snowpark.types import StructType
from snowflake.snowpark import context
from typing import Any, Union, List, Optional
from snowflake.snowpark.functions import lit
from snowflake.snowpark.dataframe import _generate_prefix

import os
import logging
import pandas as pd
from snowflake.snowpark._internal.utils import TempObjectType, random_name_for_temp_object
import re
from snowflake.snowpark.context import get_active_session

if not hasattr(DataFrameReader,"___extended_csv"):
    import logging    
    DataFrameReader.___extended_csv = True
    def _ingest_csv(self, stage_name: str, warehouse: str = None, min_warehouse_size:str = "XSMALL", max_warehouse_size:str = "XXXLARGE" ) -> str:
        if warehouse is None:
            warehouse = get_active_session().get_current_warehouse()
        pattern = self._cur_options.get("PATTERN","*.csv.*")
        quote   =   self._cur_options.get("FIELD_OPTIONALLY_ENCLOSED_BY",'"')
        sep     =   self._cur_options.get("FIELD_DELIMITER",",")
        # validate stage name
        stage_name = stage_name.strip()
        if not stage_name:
            raise Exception("PLEASE PROVIDE A NON-EMPTY STAGE NAME")
        if stage_name.startswith('@'):
            stage_name = stage_name[1:]
        session = context.get_active_session()
        # Set the warehouse
        session.sql(f"USE WAREHOUSE {warehouse}").collect()
        infer_format = random_name_for_temp_object(TempObjectType.FILE_FORMAT)
        read_format  = random_name_for_temp_object(TempObjectType.FILE_FORMAT)
        table_name   = random_name_for_temp_object(TempObjectType.TABLE)
        adjust_name  = random_name_for_temp_object(TempObjectType.FUNCTION)
        DOLLAR_DOLLAR = "$" + "$"
        # Adjust the schema of the inferred table
        session.sql(f"""
        CREATE OR REPLACE TEMP FUNCTION {adjust_name}(INPUT_ARRAY ARRAY)
        RETURNS ARRAY
        LANGUAGE JAVASCRIPT
        AS
        '
            return INPUT_ARRAY.map(row => {{
                if (row.TYPE.startsWith("NUMBER"))
                {{
                    row.EXPRESSION=`{DOLLAR_DOLLAR}{{row.ORDER_ID + 1}}::DOUBLE`;
                    row.TYPE = `DOUBLE`;
                }}
                return row;
            }});
        ';
            """).collect()

        # file format for inference
        session.sql(f"""
            CREATE OR REPLACE FILE FORMAT {infer_format} 
            TYPE = CSV 
            PARSE_HEADER=TRUE 
            FIELD_DELIMITER = ','
            FIELD_OPTIONALLY_ENCLOSED_BY = '"'
            COMPRESSION = GZIP
            """).collect()
        # file format for reading
        session.sql(f"""
            CREATE OR REPLACE FILE FORMAT {read_format}
            TYPE = CSV
            SKIP_HEADER=1
            FIELD_DELIMITER = ','
            FIELD_OPTIONALLY_ENCLOSED_BY = '"'
            COMPRESSION = GZIP;
        """).collect()
        logging.info("File formats created successfully.")
        # List csv files in the stage
        # we will use the first 5 files to infer the schema
        files = [os.path.basename(x[0]) for x in session.sql(f"""list @{stage_name} pattern='.*csv.gz'""").collect()]
        subset = files[:5]
        files_str = ",".join([f"'{x}'" for x in subset ])
        # Create a temporary table to hold the inferred schema
        session.sql(f"""
            CREATE OR REPLACE TEMP TABLE {table_name}
            USING TEMPLATE (
                SELECT {adjust_name}(ARRAY_AGG(OBJECT_CONSTRUCT(*)))
                FROM TABLE(
                    INFER_SCHEMA(
                        LOCATION=>'@{stage_name}',
                        FILE_FORMAT=>'{infer_format}',
                        FILES=>({files_str}),
                        max_records_per_file=>1000
                    )
                )
            );
        """).collect()
        logging.info("Temporary table created successfully.")
        # Get the schema of the inferred table
        describe_df = session.sql(f"describe table {table_name}")
        # We will use the schema to create a select statement
        # to read the data from the stage and insert it into the inferred table
        # We perform the following :
        # 1. We try several timestamp formats for the TIMESTAMP/DATE columns
        # 3. We try to convert BOOLEAN columns to BOOLEAN
        # 4. We try to convert NUMBER columns to NUMBER
        # 5. We try to convert DECIMAL columns to DECIMAL
        select_clauses = []
        describe_df = pd.DataFrame([row.as_dict() for row in describe_df.collect()])
        for idx, row in enumerate(describe_df.itertuples(index=False)):
            col_name = row.name
            type_str = row.type.upper()
            col_expr = f"${idx+1}::STRING"

            # Handle TIMESTAMP
            if "TIMESTAMP" in type_str:
                col_expr = (
                    f"COALESCE("
                    f"TRY_TO_TIMESTAMP({col_expr}, 'YYYY-MM-DD\"T\"HH24:MI:SS\"Z\"'), "
                    f"TRY_TO_TIMESTAMP({col_expr}, 'YYYY-MM-DD\"T\"HH24:MI:SS.FF3\"Z\"')"
                    f")"
                )
            # Handle BOOLEAN
            elif "FLOAT" in type_str:
                col_expr = f"TRY_TO_DOUBLE({col_expr})"
            # Handle BOOLEAN
            elif "BOOLEAN" in type_str:
                col_expr = f"TRY_TO_BOOLEAN({col_expr})"
            # Handle NUMBER and DECIMAL
            elif match := re.match(r"NUMBER\((\d+),(\d+)\)", type_str):
                precision, scale = match.groups()
                col_expr = f"TRY_TO_NUMBER({col_expr}, {precision}, {scale})"
            elif "NUMBER" in type_str:
                col_expr = f"TRY_TO_NUMBER({col_expr})"
            # else: leave as STRING

            select_clauses.append(f"{col_expr} AS \"{col_name}\"")

        # Replace 'INFERRED_TABLE' and 'stage_name' with actual values
        select_statement = f"INSERT INTO {table_name} SELECT\n  " + ",\n  ".join(select_clauses)
        select_statement += f"\nFROM @{stage_name} (FILE_FORMAT=>'{read_format}', PATTERN=>'.*csv.gz');"

        # Now let's scale up the warehouse to handle the data
        session.sql(f"ALTER WAREHOUSE {warehouse} SET WAREHOUSE_SIZE = '{max_warehouse_size}'").collect()

        # Execute
        try:
            session.sql(select_statement).collect()
            logging.info("Data ingested successfully.")
            logging.info(f"Files ingested: {len(files)}")
        finally:
            # Scale down the warehouse
            session.sql(f"ALTER WAREHOUSE {warehouse} SET WAREHOUSE_SIZE = '{min_warehouse_size}'").collect()
            logging.info("Warehouse scaled down successfully.")
        return session.table(table_name)

    DataFrameReader.ingest_csv = _ingest_csv


from snowflake.snowpark import DataFrame, Row, DataFrameReader
from snowflake.snowpark.types import StructType
from snowflake.snowpark import context
from typing import Any, Union, List, Optional
from snowflake.snowpark.functions import lit
from snowflake.snowpark.dataframe import _generate_prefix

if not hasattr(DataFrameReader,"___extended"):
    import logging    
    DataFrameReader.___extended = True
    DataFrameReader.__option = DataFrameReader.option
    def _option(self, key: str, value: Any, _emit_ast: bool = True) -> "DataFrameReader":
        key = key.upper()
        if key == "SEP" or key == "DELIMITER":
            key = "FIELD_DELIMITER"
        elif key == "HEADER":
            key  ="SKIP_HEADER"
            value = 1 if value == True or str(value).upper() == "TRUE" else 0
        elif key == "LINESEP":
            key  ="RECORD_DELIMITER"
        elif key == "PATHGLOBFILTER":
            key = "PATTERN"
        elif key == "CODEC":
            key = "COMPRESSION"
        elif key == "QUOTE":
            key = "FIELD_OPTIONALLY_ENCLOSED_BY"
        elif key == "NULLVALUE":
            key = "NULL_IF"
        elif key == "DATEFORMAT":
            key = "DATE_FORMAT"
        elif key == "TIMESTAMPFORMAT":
            key = "TIMESTAMP_FORMAT"
        elif key == "INFERSCHEMA":
            key = "INFER_SCHEMA"
        elif key in ["RECURSIVEFILELOOKUP","QUOTEALL","MODIFIEDBEFORE","MODIFIEDAFTER","MULTILINE","MERGESCHEMA"]:
            logging.error(f"DataFrameReader option {key} is not supported")
            return self       
        return self.__option(key,value, _emit_ast=_emit_ast)

    def _load(self,path: Union[str, List[str], None] = None, format: Optional[str] = None, schema: Union[StructType, str, None] = None,stage=None, **options) -> "DataFrame":
        self.options(dict(options))
        if "INFER_SCHEMA" in options:
            if "SKIP_HEADER" in options:
                if options["SKIP_HEADER"] == 0:
                    options["PARSE_HEADER"] = True
                else:
                    options["PARSE_HEADER"] = False
                del options["SKIP_HEADER"]
        self.format(format)
        if schema:
            self.schema(schema)
        files = []
        if isinstance(path,list):
            files.extend(path)
        else:
            files.append(path)
        session = context.get_active_session()
        if stage is None:
            stage = f'{session.get_fully_qualified_current_schema()}.{_generate_prefix("TEMP_STAGE")}'
            session.sql(f'create TEMPORARY stage if not exists {stage}').show()
        stage_files = [x for x in path if x.startswith("@")]
        if len(stage_files) > 1:
            raise Exception("Currently only one staged file can be specified. You can use a pattern if you want to specify several files")
        print(f"Uploading files using stage {stage}")
        for file in files:
            if file.startswith("file://"): # upload local file
                session.file.put(file,stage)
            elif file.startswith("@"): #ignore it is on an stage
                return self._read_semi_structured_file(file,format)
            else: #assume it is file too
                session.file.put(f"file://{file}",f"@{stage}")
        if self._file_type == "csv":
            return self.csv(f"@{stage}")
        return self._read_semi_structured_file(f"@{stage}",format)

    def _format(self, file_type: str) -> "DataFrameReader":
        file_type = str(file_type).lower()
        if file_type in ["csv","json","avro","orc","parquet","xml"]:
            self._file_type = file_type
        else:
            raise Exception(f"Unsupported file format {file_type}")
    
    DataFrameReader.format = _format
    DataFrameReader.load   = _load
    DataFrameReader.option = _option


"""Provides Additional Extensions for Snowpark"""


from .dataframe_extensions import *
from .dataframe_reader_extensions import *
from .dataframe_reader_extensions_csv import *
from .functions_extensions import *
from .session_extensions import *
from .session_builder_extensions import *
from .types_extensions import *
from .logging_utils import logged, Tag



def register_sql_magic():
    try:
        from IPython.core.magic import register_cell_magic
        def sql(line, cell):
            import IPython
            import re
            user_ns = IPython.get_ipython().user_ns
            if "session" in user_ns:
                session = user_ns['session']
                from jinja2 import Template
                t = Template(cell)
                res = t.render(user_ns)
                name = None
                if line and line.strip():
                    name = line.strip().split(" ")[0]
                # If there are several statements only last will be returned
                # also we will remove all ; at the end to avoid issues with empty statements
                res = re.sub(r';+$', '', res)
                for cursor in session.connection.execute_string(res):
                    df = session.sql(f"SELECT * FROM TABLE(RESULT_SCAN('{cursor.sfqid}'))")
                    # to avoid needed to do a count on display
                    setattr(df,"_cached_rowcount",cursor.rowcount)
                if name:
                    user_ns[name] = df
                else:
                    user_ns["__df"] = df
                    return df
            else:
                return "No session was found. You can setup one by running: session = Session.builder.from_env().getOrCreate()"
        register_cell_magic(sql)
    except:
        pass

register_sql_magic()



import pytest
import snowpark_extensions
from snowflake.snowpark import Session
from snowflake.snowpark.types import *
from snowflake.snowpark.functions import col,lit, array_sort,sort_array, array_max, array_min, map_values, struct,object_construct, array_agg, bround
from snowflake.snowpark import functions as F
import re

def test_asc():
    session = Session.builder.from_snowsql().getOrCreate()
    df = session.createDataFrame([('Tom', 80), ('Alice', None)], ["name", "height"])
    res = df.select(df.name).orderBy(F.asc(df.name)).collect()
    assert res[0].NAME == 'Alice'

def test_desc():
    session = Session.builder.from_snowsql().getOrCreate()
    df = session.createDataFrame([('Tom', 80), ('Alice', None)], ["name", "height"])
    res = df.select(df.name).orderBy(F.desc(df.name)).collect()
    assert res[0].NAME == 'Tom'

def test_array():
    session = Session.builder.from_snowsql().getOrCreate()
    df = session.sql("select 1 A")
    res=df.withColumn("array",F.array(lit(1),lit(2),lit(3))).collect()
    assert len(res)==1
    array = eval(res[0][1])
    assert array[0]==1 and array[1]==2 and array[2]==3

def test_array_distinct():
    session = Session.builder.from_snowsql().getOrCreate()
    df = session.sql("select 1 A")
    df=df.withColumn("array",F.array(lit(1),lit(1),lit(1),lit(2),lit(3),lit(2),lit(2)))
    res=df.withColumn("array_d",F.array_distinct("ARRAY")).collect()
    assert len(res)==1
    array = eval(res[0][2])
    assert len(array)==3
    assert array[0]==1 and array[1]==2 and array[2]==3

def test_array_flatten():
    session = Session.builder.from_snowsql().getOrCreate()
    df = session.createDataFrame([([[1, 2, 3], [4, 5], [6]],1), ([[1],None, [4, 5]],2)], ['data','pos'])
    res=df.select(F.flatten(df.data).alias("FLATTEN"))
    res=res.collect()
    assert len(res)==2
    array1 = eval(res[0]['FLATTEN'])
    array2 = eval(res[1]['FLATTEN'].replace("null",'None'))
    assert array1[0]==1 and array1[1]==2 and array1[2]==3 and array1[3]==4 and array1[4]==5 and array1[5]==6
    assert array2 == [1, None, 4, 5] 

def test_create_map():
    def do_assert(res):
        assert len(res) == 5
        assert res[0].ID == '34534' and res[0].DEPT == "Sales"     and eval(res[0].PROPERTIESMAP)['location'] == 'USA' and eval(res[0].PROPERTIESMAP)['salary'] == 6500
        assert res[1].ID == '36636' and res[1].DEPT == "Finance"   and eval(res[1].PROPERTIESMAP)['location'] == 'USA' and eval(res[1].PROPERTIESMAP)['salary'] == 3000
        assert res[2].ID == '39192' and res[2].DEPT == "Marketing" and eval(res[2].PROPERTIESMAP)['location'] == 'CAN' and eval(res[2].PROPERTIESMAP)['salary'] == 2500
        assert res[3].ID == '40288' and res[3].DEPT == "Finance"   and eval(res[3].PROPERTIESMAP)['location'] == 'IND' and eval(res[3].PROPERTIESMAP)['salary'] == 5000
        assert res[4].ID == '42114' and res[4].DEPT == "Sales"     and eval(res[4].PROPERTIESMAP)['location'] == 'USA' and eval(res[4].PROPERTIESMAP)['salary'] == 3900
    session = Session.builder.app_name('test').from_snowsql().getOrCreate()
    data = [ ("36636","Finance",3000,"USA"), 
        ("40288","Finance",5000,"IND"), 
        ("42114","Sales",3900,"USA"), 
        ("39192","Marketing",2500,"CAN"), 
        ("34534","Sales",6500,"USA") ]
    schema = StructType([
        StructField('id', StringType(), True),
        StructField('dept', StringType(), True),
        StructField('salary', IntegerType(), True),
        StructField('location', StringType(), True)
        ])
    df = session.createDataFrame(data=data,schema=schema)
    df.show()
    #Convert columns to Map
    from snowflake.snowpark.functions import col,lit,create_map
    res = df.withColumn("propertiesMap",create_map(
            lit("salary"),col("salary"),
            lit("location"),col("location")
            )).drop("salary","location").sort("ID").collect()
    do_assert(res)
    res=df.withColumn("propertiesMap",create_map([
            lit("salary"),col("salary"),
            lit("location"),col("location")]
            )).drop("salary","location").sort("ID").collect()
    do_assert(res)
    res=df.withColumn("propertiesMap",create_map((
            lit("salary"),col("salary"),
            lit("location"),col("location"))
            )).drop("salary","location").sort("ID").collect()
    do_assert(res)
    res=df.withColumn("propertiesMap",create_map((
            (lit("salary"),col("salary")),
            (lit("location"),col("location"))
            ))).drop("salary","location").sort("ID").collect()
    do_assert(res)


def test_map_values():
    session = Session.builder.from_snowsql().getOrCreate()
    df = session.sql("SELECT object_construct('1', 'a', '2', 'b') as data")
    res = df.select(map_values("data").alias("values")).collect()
    # +------+
    # |values|
    # +------+
    # |[a, b]|
    # +------+
    assert len(res)==1
    array=re.sub(r"\s","",res[0].VALUES)
    assert array == '["a","b"]'
    df = session.sql("SELECT object_construct('1', 'value1', '2', parse_json('null')) as data")
    res = df.select(map_values("data").alias("values")).collect()
    assert len(res)==1
    array=re.sub(r"\s","",res[0].VALUES)
    assert array == '["value1",null]'

def test_struct():
    session = Session.builder.from_snowsql().getOrCreate()
    df = session.createDataFrame([('Bob', 80), ('Alice', None)], ["name", "age"])
    res=df.select(struct('age', 'name').alias("struct")).collect()    
#     [Row(struct=Row(age=2, name='Alice')), Row(struct=Row(age=5, name='Bob'))]
    assert len(res)==2
    assert re.sub(r"\s","",res[0].STRUCT) == '{"age":80,"name":"Bob"}'
    assert re.sub(r"\s","",res[1].STRUCT) == '{"age":null,"name":"Alice"}'
    res = df.select(struct([df.age, df.name]).alias("struct")).collect()
    assert len(res)==2
    assert re.sub(r"\s","",res[0].STRUCT) == '{"AGE":80,"NAME":"Bob"}'
    assert re.sub(r"\s","",res[1].STRUCT) == '{"AGE":null,"NAME":"Alice"}'    
#     [Row(struct=Row(age=2, name='Alice')), Row(struct=Row(age=5, name='Bob'))]
    res = df.select(struct(df.age.alias("A"), df.name.alias("B")).alias("struct")).collect()
    assert len(res)==2
    assert re.sub(r"\s","",res[0].STRUCT) == '{"A":80,"B":"Bob"}'
    assert re.sub(r"\s","",res[1].STRUCT) == '{"A":null,"B":"Alice"}'

def test_daydiff():
    session = Session.builder.from_snowsql().getOrCreate()
    df = session.createDataFrame([('2015-04-08','2015-05-10')], ['d1', 'd2'])
    res = df.select(F.daydiff(F.to_date(df.d2), F.to_date(df.d1)).alias('diff')).collect()
    assert res[0].DIFF == 32  


def test_bround():
    session = Session.builder.from_snowsql().getOrCreate()
    data0 = [(1.5,0),
    (2.5,0),
    (0.00,0),
    (0.5,0),
    (-1.5,0),
    (-2.5,0)]

    data1 = [
    (2.25,1),
    (2.65,1),
    (0.00,1),
    (1.05,1),
    (1.15,1),
    (-2.25,1),
    (-2.35,1),
    (None,1),
    (1.5,1),
    (1.5,-1) ]

    data_null = [
    (0.5,None),
    (1.5,None),
    (2.5,None),
    (-1.5,None),
    (-2.5,None),
    (None,None)]
    schema_df = StructType([
    StructField('value', FloatType(), True),
    StructField('scale', IntegerType(), True)
    ])
    
    df_0 = session.createDataFrame(data0, schema_df)
    df_1 = session.createDataFrame(data1, schema_df)
    df_null = session.createDataFrame(data_null, schema_df)

    res0 = df_0.withColumn("rounding",F.bround(F.col('value').cast(DecimalType(20,4)),0)).collect()
    assert len(res0) == 6
    assert res0[0].ROUNDING == 2.0
    assert res0[1].ROUNDING == 2.0
    assert res0[2].ROUNDING == 0.0
    assert res0[3].ROUNDING == 0.0
    assert res0[4].ROUNDING == -2.0
    assert res0[5].ROUNDING == -2.0

    res1 = df_1.withColumn("rounding",F.bround(F.col('value').cast(DecimalType(20,4)),1)).collect()
    assert len(res1) == 10
    assert str(res1[0].ROUNDING) == "2.2"
    assert str(res1[1].ROUNDING) == "2.6"
    assert str(res1[2].ROUNDING) == "0.0"
    assert str(res1[3].ROUNDING) == "1.0"
    assert str(res1[4].ROUNDING) == "1.2"
    assert str(res1[5].ROUNDING) == "-2.2"
    assert str(res1[6].ROUNDING) == "-2.4"
    assert str(res1[7].ROUNDING) == "None"
    assert str(res1[8].ROUNDING) == "1.5"
    assert str(res1[9].ROUNDING) == "1.5"

    resNull = df_null.withColumn("rounding",F.bround(F.col('value').cast(DecimalType(20,4)),None) ).collect()
    assert len(resNull) == 6
    assert str(resNull[0].ROUNDING) == "None"
    assert str(resNull[1].ROUNDING) == "None"
    assert str(resNull[2].ROUNDING) == "None"
    assert str(resNull[3].ROUNDING) == "None"
    assert str(resNull[4].ROUNDING) == "None"
    assert str(resNull[5].ROUNDING) == "None"

def test_regexp_split():
    session = Session.builder.from_snowsql().config("schema","PUBLIC").getOrCreate()
    from snowflake.snowpark.functions import regexp_split
    df = session.createDataFrame([('testAandtestBareTwoBBtests',)], ['s',])
    res = df.select(regexp_split(df.s, "test(A|BB)" , 3).alias('s')).collect()
    assert res[0].S == '[\n  "",\n  "andtestBareTwoBBtests"\n]'
    res = df.select(regexp_split(df.s, "test(A|BB)", 1).alias('s')).collect()
    assert res[0].S == '[\n  "testAandtestBareTwoBBtests"\n]'

    df = session.createDataFrame([('From: mauricio@mobilize.net',)], ['s',])

    res = df.select(regexp_split(df.s, "((From|To)|Subject): (\w+@\w+\.[a-z]+)").alias('s')).collect()
    assert res[0].S == '[\n  "",\n  ""\n]'

    df = session.createDataFrame([('oneAtwoBthreeC',)], ['s',])
    
    res = df.select(regexp_split(df.s, 'Z').alias('s')).collect()
    assert res[0].S == '[\n  "oneAtwoBthreeC"\n]'
    res = df.select(regexp_split(df.s, 't').alias('s')).collect()
    assert res[0].S == '[\n  "oneA",\n  "woB",\n  "hreeC"\n]'
    res = df.select(regexp_split(df.s, 't', 1).alias('s')).collect()
    assert res[0].S == '[\n  "oneAtwoBthreeC"\n]'
    res = df.select(regexp_split(df.s, 't', 2).alias('s')).collect()
    assert res[0].S == '[\n  "oneA",\n  "woBthreeC"\n]'    
    res = df.select(regexp_split(df.s, '[ABC]').alias('s')).collect()
    assert res[0].S == '[\n  "one",\n  "two",\n  "three",\n  ""\n]'    
    res = df.select(regexp_split(df.s, '[ABC]', 1).alias('s')).collect()
    assert res[0].S == '[\n  "oneAtwoBthreeC"\n]'  
    res = df.select(regexp_split(df.s, '[ABC]', 2).alias('s')).collect()
    assert res[0].S == '[\n  "one",\n  "twoBthreeC"\n]'   
    
    df = session.createDataFrame([('HelloabNewacWorld',)], ['s',])

    res = df.select(regexp_split(df.s, 'a([b, c]).*?').alias('s')).collect()
    assert res[0].S == '[\n  "Hello",\n  "New",\n  "World"\n]'

    df = session.createDataFrame([(r'aa\nbb\nccc\b',)], ['s',])
    
    res = df.select(regexp_split(df.s, r'\w+.').alias('s')).collect()
    assert res[0].S == '[\n  "",\n  "",\n  "",\n  "b"\n]'

    df = session.createDataFrame([(r'\n\n\n',)], ['s',])   

    res = df.select(regexp_split(df.s, '.*', 3).alias('s')).collect()
    assert res[0].S == '[\n  "",\n  "",\n  ""\n]'

    df = session.createDataFrame([("""line 1
line 2
line 3""",)], ['s',])    

    res = df.select(regexp_split(df.s, r'\n', 3).alias('s')).collect()
    assert res[0].S == '[\n  "line 1",\n  "line 2",\n  "line 3"\n]'
    res = df.select(regexp_split(df.s, r'line 1(\n)', 3).alias('s')).collect()
    assert res[0].S == '[\n  "",\n  "line 2\\nline 3"\n]'

    df = session.createDataFrame([('The price of PINEAPPLE ice cream is 20',)], ['s',])
    res = df.select(regexp_split(df.s, r"(\b[A-Z]+\b).+(\b\d+)", 4).alias('s')).collect()
    assert res[0].S == '[\n  "The price of ",\n  ""\n]'

    df = session.createDataFrame([('<button type="submit" class="btn">Send</button>',)], ['s',])
    res = df.select(regexp_split(df.s, '".+?"', 4).alias('s')).collect()
    assert res[0].S == '[\n  "<button type=",\n  " class=",\n  ">Send</button>"\n]'

def test_to_utc_timestamp_ext():
    session = Session.builder.from_snowsql().config("schema","PUBLIC").getOrCreate()
    from snowflake.snowpark.functions import to_utc_timestamp_ext
    df = session.createDataFrame([('1997-02-28 10:30:00', 'JST')], ['ts', 'tz'])
    res = df.select(F.to_utc_timestamp_ext(df.ts, "PST").alias('utc_time')).collect()
    assert res[0][0] == datetime.datetime(1997, 2, 28, 18, 30)
    res = df.select(F.to_utc_timestamp_ext(df.ts, df.tz).alias('utc_time')).collect()
    assert res[0][0] == datetime.datetime(1997, 2, 28, 1, 30)



import pytest
import snowpark_extensions
from snowflake.snowpark import Session
from snowflake.snowpark.types import *
from snowflake.snowpark.functions import col,lit, array_sort,sort_array, array_max, array_min, map_values, struct,object_construct, array_agg, bround
from snowflake.snowpark import functions as F
import re

#def test_ingest_csv():
session = Session.builder.from_snowsql().getOrCreate()
df = session.read.option("delimiter",",").ingest_csv("SNOWPARK_TESTDB.PUBLIC.BUG1","XSMALL_WH")
df.show()





import pytest
from snowflake.snowpark import Session, Row
from snowflake.snowpark.types import *
import snowpark_extensions

def test_load():    
    session = Session.builder.from_snowsql().getOrCreate()
    schema = StructType([ \
    StructField("case_id",       StringType()), \
    StructField("province",      StringType()), \
    StructField("city",          StringType()), \
    StructField("group",         BooleanType()), \
    StructField("infection_case",StringType()), \
    StructField("confirmed",     IntegerType()), \
    StructField("latitude",      FloatType()), \
    StructField("cilongitudety", FloatType()) \
    ])
    cases = session.read.load(["./tests/data/test1_0.csv","./tests/data/test1_1.csv"],
                        schema=schema,
                        format="csv", 
                        sep=",",
                        header="true")
    assert 10 == len(cases.collect())


from snowflake.snowpark.types import StructType, StructField

if not hasattr(StructType, "___extended"):
  StructType.___extended = True    
  def schema__get_item__(self,index):
      if isinstance(index, str):
         name = index.upper()
         for f in self.fields:
            if name == f.name.upper():
               return f
      else:
         return self.fields[index]
  StructType.__getitem__ = schema__get_item__
  StructField.dtype = property(lambda self: self.datatype)
  from snowflake.snowpark._internal.type_utils import convert_sf_to_sp_type
  def fromJson(cls, json: dict) -> StructField:
        return StructField(
            json["name"],
            convert_sf_to_sp_type(json["type"]),
            json.get("nullable", True),
            json.get("metadata"),
        )
  StructField.fromJson = fromJson
  StructType.fromJson =  lambda cls,json: StructType([StructField.fromJson(f) for f in json["fields"]])


def logged(procname):
    import logging, datetime
    from functools import wraps
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            log_filename = f"/tmp/{procname}_{datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}.log"
            logging.basicConfig(filename=log_filename, level=logging.DEBUG)
            try:
                result = func(*args, **kwargs)
            except Exception as ex:
                logging.exception(ex)
                result = "error"
            finally:
                logging.shutdown()
                session = args[0]
                session.file.put(f'file://{log_filename}','@~/')
            return result
        return wrapper
"""    
 EXAMPLE:

 Logging Decorator can be used like this:
 -------------------------------------------
 create or replace procedure logging_test()
 returns variant
 language python
 runtime_version=3.8
 packages = ('snowflake-snowpark-python')
 handler='process'
 as $$
 def logged(procname):
     import logging, datetime
     from functools import wraps
     def decorator(func):
         @wraps(func)
         def wrapper(*args, **kwargs):
             log_filename = f"/tmp/{procname}_{datetime.datetime.now().strftime('%Y-%m-%d-%H-%M-%S')}.log"
             logging.basicConfig(filename=log_filename, level=logging.DEBUG)
             try:
                 result = func(*args, **kwargs)
             except Exception as ex:
                 logging.exception(ex)
                 result = "error"
             finally:
                 logging.shutdown()
                 session = args[0]
                 session.file.put(f'file://{log_filename}','@~/')
             return result
         return wrapper
     return decorator
 @logged('logging_test')    
 def process(session,*args):
     import logging
     logger = logging.getLogger('process') # it will be great to have an env_var with the current proc name
     logger.setLevel(logging.DEBUG)
     logger.info("info1")
     logger.info("info1")
     logger.warning("warning")
     logger.error("error")
     session.sql("select count(*) from information_schema.tables")
     logger.error("error2")
     return "Done"
 $$;
 -- to call it use:
 call logging_test()
 -- to see traces just do:
 ls @~ PATTERN = 'logging_test.*';
 -- see the traces with
 select $1 from @~/logging_test_2023-02-14-09-07-18.log.gz
 -- delete the traces
 remove @~ PATTERN = 'logging_test.*';
 -- NOTE: for some reason I can not get the info or warning traces.
"""

"""
 this code has been borrow from: 
 Simple Tags in Snowflake Snowpark for Python
 https://medium.com/snowflake/simple-tags-in-snowflake-snowpark-for-python-c5910749273
 Special thanks to Bart
"""
from contextlib import ContextDecorator
class Tag(ContextDecorator):
    def __init__(self, session, tag_name, label=None):
        self.old_tag = None
        self.session = session
        self.tag_name = tag_name
        self.label = None
    def __call__(self, func):
        if self.label is None:  # Label was not provided
            self.label = func.__name__  # Use function's name.
        return super().__call__(func)
    def __enter__(self):
        self.old_tag = self.session.query_tag
        if self.label:
            self.session.query_tag=f"{self.tag_name}_{self.label}"
        else:
            self.session.query_tag=self.tag_name
        return self
    def __exit__(self, *exc):
        self.session.query_tag=self.old_tag
        return False


from setuptools import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

VERSION = '0.0.47'

setup(name='snowpark_extensions',
      version=VERSION,
      description='A set of helpers to extend snowpark functionality. NOTE: most of this extensions are already available in the official snowpark python package',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/Snowflake-Labs/snowpark-extensions-py',
      author='mauricio.rojas',
      install_requires=['snowflake-snowpark-python[pandas]>=1.31.1','jinja2'],
      author_email='mauricio.rojas@snowflake.com',
      packages=['snowpark_extensions'],
      zip_safe=False)



from snowflake.snowpark import DataFrame, Row, DataFrameNaFunctions, Column
from snowflake.snowpark.functions import col, lit, udtf, regexp_replace
from snowflake.snowpark import functions as F
from snowflake.snowpark.dataframe import _generate_prefix
from snowflake.snowpark.column import _to_col_if_str
import pandas as pd
from snowpark_extensions.utils import schema_str_to_schema
from snowflake.snowpark.types import StructType,StructField
from snowflake.snowpark._internal.analyzer.expression import FunctionExpression, UnresolvedAttribute
from snowflake.snowpark._internal.analyzer.unary_expression import Alias
from snowflake.snowpark._internal.analyzer.analyzer_utils import quote_name
from snowflake.snowpark import Column
from snowflake.snowpark.types import *
from snowflake.snowpark.functions import udtf, col, hash as sf_hash
from snowflake.snowpark.relational_grouped_dataframe import RelationalGroupedDataFrame

from typing import (
    Dict,
    Iterable,
    Optional,
    Union,
    List
)
from snowflake.snowpark._internal.type_utils import (
    ColumnOrName,
    LiteralType,
)
from snowflake.snowpark._internal.utils import generate_random_alphanumeric

from snowflake.snowpark.table_function import (
    TableFunctionCall,
)



if not hasattr(DataFrame,"___extended"):
    
    DataFrame.___extended = True

    def _repr_html_(self):
        import IPython
        rows_limit = getattr(DataFrame,'__rows_limit',50)
        if 'display' in globals():
            display = globals()['display']
        elif 'display' in IPython.get_ipython().user_ns:
            display = IPython.get_ipython().user_ns['display']
        else:
            from IPython.display import display
        try:
            data_to_display = self.limit(rows_limit).collect()
            if len(data_to_display) == 0:
                return "No rows to display"
            else:
                df = pd.DataFrame.from_records([x.as_dict() for x in data_to_display])
                if len(data_to_display) >= rows_limit:
                    print(f"Showing only {rows_limit}. Change DataFrame.__rows_limit value to display more rows")
                display(df)
            return ""
        except Exception as ex:
                return str(ex)

    setattr(DataFrame,'_repr_html_',_repr_html_)
    setattr(DataFrame,'__rows_limit',50)

    def _ipython_key_completions_(self) -> List[str]:
        """Returns the names of columns in this :class:`DataFrame`.
        """
        return self.columns
    setattr(DataFrame,'_ipython_key_completions_',_ipython_key_completions_)

    def map(self,func,output_types,input_types=None,input_cols=None,to_row=False):
        clazz= _generate_prefix("map")
        output_schema=[]
        if not input_types:
            input_types = [x.datatype for x in self.schema.fields]
        input_cols_len=len(input_types)
        if not input_cols:
            input_col_names=self.columns[:input_cols_len]
            _input_cols = [self[x] for x in input_col_names]
        else:
            input_col_names=input_cols
            _input_cols = input_cols
        output_schema = [StructField(f"c_{i+1}",output_types[i]) for i in range(len(output_types))]
        output_cols = [col(x.name) for x in output_schema]
        if to_row:
            myRow = Row(*input_col_names)
            def process_func(self,*argv):
                yield func(myRow(*argv))
        else:
            def process_func(self,*argv):
                yield func(*argv)
        udtf_class = type(clazz, (object, ), {"process":process_func})
        udtf(udtf_class,output_schema=StructType(output_schema), input_types=input_types,name=clazz,replace=True,packages=["snowflake-snowpark-python"])
        return self.join_table_function(clazz,*_input_cols).select(*output_cols)

    DataFrame.map = map

    def simple_map(self,func):
        return self.select(*func(self))

    DataFrame.simple_map = simple_map

    DataFrameNaFunctions.__oldreplace = DataFrameNaFunctions.replace

    def extended_replace(
        self,
        to_replace: Union[
            LiteralType,
            Iterable[LiteralType],
            Dict[LiteralType, LiteralType],
        ],
        value: Optional[Iterable[LiteralType]] = None,
        subset: Optional[Iterable[str]] = None,
        regex=False
    ) -> "snowflake.snowpark.dataframe.DataFrame":
        if regex:
            return self._df.select([regexp_replace(col(x.name), to_replace,value).alias(x.name) if isinstance(x.datatype,StringType) else col(x.name) for x in self._df.schema])
        else:
            return self.__oldreplace(to_replace,value,subset)

    DataFrameNaFunctions.replace = extended_replace

    def has_null(col):
        return F.array_contains(F.sql_expr("parse_json('null')"),col) | F.coalesce(F.array_contains(lit(None) ,col),lit(False))

    def stack(self,rows:int,*cols):
        count_cols = len(cols)
        if count_cols % rows != 0:
            raise Exception("Invalid parameter. The given cols cannot be arrange in the give rows")
        out_count_cols = int(count_cols / rows)
        # determine the input schema
        input_schema = self.select(cols).limit(1).schema
        from snowflake.snowpark.functions import col
        input_types = [x.datatype for x in input_schema.fields]
        input_cols  = [x.name     for x in input_schema.fields]
        output_cols = [f"col{x}" for x in range(1,out_count_cols)]
        clazz=_generate_prefix("stack")
        def process(self, *row):
            for i in range(0, len(row), out_count_cols):
                yield tuple(row[i:i+out_count_cols])
        output_schema = StructType([StructField(f"col{i+1}",input_schema.fields[i].datatype) for i in range(0,out_count_cols)])
        udtf_class = type(clazz, (object, ), {"process":process})
        tfunc = udtf(udtf_class,output_schema=output_schema, input_types=input_types,name=clazz,replace=True,is_permanent=False,packages=["snowflake-snowpark-python"])
        return tfunc(*cols)
    
    DataFrame.stack = stack

    class GroupByPivot():
      def __init__(self,old_groupby_col,pivot_col):
            self.old_groupby_col = old_groupby_col
            pivot_col = _to_col_if_str(pivot_col,"pivot")
            self.pivot_col=pivot_col
            group_by_exprs = [F.sql_expr(x.sql) for x in old_groupby_col._grouping_exprs]
            group_by_exprs.append(pivot_col)
            # version of snowpark < 1.26.0 stored the dataframe in _df, while later versions store it in _dataframe
            old_dataframe_ref = getattr(old_groupby_col,"_df",None) or getattr(old_groupby_col,"_dataframe",None)
            if old_dataframe_ref is None:
                raise Exception("Cannot find dataframe reference")
            self.df = old_dataframe_ref.groupBy(group_by_exprs)
      def clean(self,pivoted):
        def get_valid_id(id):
            return id if re.match("[A-Za-z]\w+", id) else f'"{id}"'
        n = len(self.value_list)
        last_n_cols = self.value_list[-n:]
        pivoted_cols = pivoted.columns[-n:]
        previous_cols = pivoted.columns[:len(pivoted.columns)-n]
        for i in range(0,n):
            the_col = pivoted_cols[i]
            the_value = self.value_list[i]
            pivoted_cols[i] = col(the_col).alias(get_valid_id(str(the_value)))
        return pivoted.select(*previous_cols,*pivoted_cols)
      def prepare(self,prepared_col):
            first = self.df.agg(prepared_col.alias("__firstAggregate"))
            self.value_list = [x[0] for x in first.select(self.pivot_col).distinct().sort(self.pivot_col).collect()]
            return first.pivot(pivot_col=self.pivot_col,values=self.value_list)
      def avg(self, col: ColumnOrName) -> DataFrame:
            """Return the average for the specified numeric columns."""
            return self.prepare(F.avg(col)).avg("__firstAggregate")
      mean = avg
      def sum(self, col: ColumnOrName) -> DataFrame:
            """Return the sum for the specified numeric columns."""
            return  self.clean(self.prepare(F.sum(col)).sum("__firstAggregate"))
      def median(self, col: ColumnOrName) -> DataFrame:
            """Return the median for the specified numeric columns."""
            return self.clean(self.prepare(F.median(col)).median("__firstAggregate"))
      def min(self, col: ColumnOrName) -> DataFrame:
            """Return the min for the specified numeric columns."""
            return self.clean(self.prepare(F.min(col)).min("__firstAggregate"))
      def max(self, col: ColumnOrName) -> DataFrame:
            """Return the max for the specified numeric columns."""
            return self.clean(self.prepare(col).max("__firstAggregate"))
      def count(self) -> DataFrame:
            """Return the number of rows for each group."""
            if hasattr(self.pivot_col, "_expression") and (isinstance(self.pivot_col._expression, UnresolvedAttribute) or isinstance(self.pivot_col,Alias)):
                colname = self.pivot_col._expression.name
                first = self.df.agg(self.pivot_col)
                self.value_list = [x[0] for x in first.select(self.pivot_col).distinct().sort(self.pivot_col).collect()]
                return self.clean(first.pivot(pivot_col=self.pivot_col,values=self.value_list).agg(F.sql_expr(f"COUNT({colname})")))
            else:
                raise Exception("pivot_col expression is not supported")
      def agg(self, aggregated_col: ColumnOrName) -> DataFrame:
            if hasattr(aggregated_col, "_expression") and isinstance(aggregated_col._expression, FunctionExpression):
                name = aggregated_col._expression.name
                return self.clean(self.prepare(aggregated_col).function(name)(col("__firstAggregate")))
            else:
                raise Exception("Alias functions expressions are supported")

    def group_by_pivot(self,pivot_col):
        return GroupByPivot(self, pivot_col)
    RelationalGroupedDataFrame.pivot = group_by_pivot

    def transform(self, func, *args, **kwargs):
            result = func(self, *args, **kwargs)
            return result

    DataFrame.transform = transform

    


  ###### HELPER END

  






from snowflake.snowpark import functions as F
from snowflake.snowpark import context
from snowflake.snowpark.functions import call_builtin, col,lit, concat, coalesce, object_construct_keep_null, table_function, udf
from snowflake.snowpark import DataFrame, Column
from snowflake.snowpark.types import ArrayType, BooleanType
from snowflake.snowpark._internal.type_utils import (
    ColumnOrLiteral,
    ColumnOrLiteralStr,
    ColumnOrName,
    ColumnOrSqlExpr,
    LiteralType,
)
from snowflake.snowpark.column import _to_col_if_str, _to_col_if_lit, _to_col_if_str_or_int
from snowflake.snowpark.dataframe import _generate_prefix
from snowflake.snowpark._internal.analyzer.unary_expression import Alias
import re

if not hasattr(F,"___extended"):
    F.___extended = True

    def pairwise(iterable):
        while len(iterable):
            a = iterable.pop(0)
            if len(iterable):
                b = iterable.pop(0)
            else:
                b = None
            yield (a,b)
    
    def flatten_col_list(obj):
        if isinstance(obj, str) or isinstance(obj, Column):
            return [obj]
        elif hasattr(obj, '__iter__'):
            acc = []
            for innerObj in obj:
                acc = acc + flatten_col_list(innerObj)
            return acc

    def format_number(col,d):
        col = _to_col_if_str(col,"format_number")
        return F.to_varchar(col,'999,999,999,999,999.' + '0'*d)

    def create_map(*col_names):
        """
        Usage:
        res = df.select(create_map('name', 'age').alias("map")).collect()
        """
        from snowflake.snowpark.functions import col,lit, object_construct
        col_list = []
        # flatten any iterables, to process them in pairs
        col_names = flatten_col_list(col_names)
        for name, value in pairwise(col_names):
            col_list.append(_to_col_if_str(name,"create_map"))
            col_list.append(value)
        return object_construct(*col_list)
    
    def _array(*cols):
        return F.array_construct(*cols)

    def _struct(*cols):
        new_cols = []
        for c in flatten_col_list(cols):
            if isinstance(c, str):
                new_cols.append(lit(c))
            else:
                name = c._expression.name
                name = name[1:] if name.startswith('"') else name
                name = name[:-1] if name.endswith('"') else name
                new_cols.append(lit(name))
            c = _to_col_if_str(c, "struct")
            if isinstance(c, Column) and isinstance(c._expression,Alias):
                new_cols.append(col(c._expression.children[0])) 
            else:
                new_cols.append(c)
        return object_construct_keep_null(*new_cols)

    F._array_flatten_udf = None
    def _array_flatten(array):
        if not F._array_flatten_udf:
            @udf
            def _array_flatten(array_in:list) -> list:
                flat_list = []
                for sublist in array_in:
                    if type(sublist) == list:
                        flat_list.extend(sublist)
                    else:
                        flat_list.append(sublist)              
                return flat_list
            F._array_flatten_udf = _array_flatten
        array = _to_col_if_str(array, "array_flatten")
        return F._array_flatten_udf(array)

    F._array_zip_udfs = {}

    def build_array_zip_ddl(nargs:int):
        function_name = _generate_prefix(f"array_zip_{nargs}")
        args          = ",".join([f"list{x} ARRAY" for x in range(1,nargs+1)])
        args_names    = ",".join([f"list{x}"       for x in range(1,nargs+1)])
        return function_name,f"""
CREATE OR REPLACE TEMPORARY FUNCTION {function_name}({args})
returns ARRAY language python runtime_version = '3.8'
handler = 'zip_list'
as
$$
def zip_list({args_names}):
    return list(zip({args_names}))
$$;"""

    def _arrays_zip(*lists):
        nargs = len(lists)
        if nargs < 2:
            raise Exception("At least two list are needed for array_zip")
        if not str(nargs) in F._array_zip_udfs:
            try:
                function_name, udf_ddl = build_array_zip_ddl(nargs)
                context.get_active_session().sql(udf_ddl).show()
                F._array_zip_udfs[str(nargs)] = function_name
            except Exception as e:
                raise Exception(f"Could not register support udf for array_zip. Error: {e}")
        list_cols = [_to_col_if_str(x, "array_zip") for x in lists]
        return F.call_builtin(F._array_zip_udfs[str(nargs)],*list_cols)

    def _bround(col: Column, scale: int = 0): 
        power = pow(F.lit(10), F.lit(scale))
        elevatedColumn = F.when(F.lit(0) == F.lit(scale), col).otherwise(col * power)
        columnFloor = F.floor(elevatedColumn)
        return F.when(
            elevatedColumn - columnFloor == F.lit(0.5)
            , F.when(columnFloor % F.lit(2) == F.lit(0), columnFloor).otherwise(columnFloor + F.lit(1))
        ).otherwise(F.round(elevatedColumn)) / F.when(F.lit(0) == F.lit(scale), F.lit(1)).otherwise(power)
    
    def has_special_char(string):
        pattern = '[^A-Za-z0-9]+'
        result = re.search(pattern, string)
        return bool(result)

    def is_not_a_regex(pattern):
        return not has_special_char(pattern)

    F._split_regex_udf = None
    def _regexp_split(value:ColumnOrName, pattern:ColumnOrLiteralStr, limit:int = -1):  
        value = _to_col_if_str(value,"split_regex")                
        pattern_col = pattern        
        if isinstance(pattern, str):
            pattern_col = lit(pattern)        
        if limit < 0 and isinstance(pattern, str) and is_not_a_regex(pattern):
            return F.split(value, pattern_col)  
                    
        session = context.get_active_session()
        current_database = session.get_current_database() 
        function_name =_generate_prefix("_regex_split_helper")           
        F._split_regex_udf = f"{current_database}.public.{function_name}"

        session.sql(f"""CREATE OR REPLACE FUNCTION {F._split_regex_udf} (input String, regex String, limit INT)
RETURNS ARRAY
LANGUAGE JAVA
RUNTIME_VERSION = '11'
PACKAGES = ('com.snowflake:snowpark:latest')
HANDLER = 'MyJavaClass.regex_split_run' 
AS
$$
import java.util.regex.Pattern;
public class MyJavaClass {{
    public String[] regex_split_run(String input,String regex, int limit) {{
        Pattern pattern = Pattern.compile(regex);
        return pattern.split(input, limit);
    }}}}$$;""").show()
        return call_builtin(F._split_regex_udf, value, pattern_col, limit)

    F._map_values_udf = None
    def _map_values(col:ColumnOrName):
        col = _to_col_if_str(col,"map_values")
        if not F._map_values_udf:
            @udf(replace=True,is_permanent=False)
            def map_values(obj:dict)->list:
                return list(obj.values())
            F._map_values_udf = map_values
        return F._map_values_udf(col)

    timezoneMap = {
        'EST': '-05:00',
        'HST': '-10:00',
        'MST': '-07:00',
        'ACT': 'Australia/Darwin',
        'AET': 'Australia/Sydney',
        'AGT': 'America/Argentina/Buenos_Aires',
        'ART': 'Africa/Cairo',
        'AST': 'America/Anchorage',
        'BET': 'America/Sao_Paulo',
        'BST': 'Asia/Dhaka',
        'CAT': 'Africa/Harare',
        'CNT': 'America/St_Johns',
        'CST': 'America/Chicago',
        'CTT': 'Asia/Shanghai',
        'EAT': 'Africa/Addis_Ababa',
        'ECT': 'Europe/Paris',
        'IET': 'America/Indiana/Indianapolis',
        'IST': 'Asia/Kolkata',
        'JST': 'Asia/Tokyo',
        'MIT': 'Pacific/Apia',
        'NET': 'Asia/Yerevan',
        'NST': 'Pacific/Auckland',
        'PLT': 'Asia/Karachi',
        'PNT': 'America/Phoenix',
        'PRT': 'America/Puerto_Rico',
        'PST': 'America/Los_Angeles',
        'SST': 'Pacific/Guadalcanal',
        'VST': 'Asia/Ho_Chi_Minh'
    }
    def map_timestamp(tz_col):
         from functools import reduce
         when_exprs = F
         for tz, offset in timezoneMap.items():
             when_exprs = when_exprs.when( tz_col == F.lit(tz), F.lit(offset))
         return when_exprs.otherwise(tz_col)
    def to_utc_timestamp_ext(timestamp:ColumnOrName, tz:ColumnOrLiteral):
        """
        The function converts a timezone-agnostic timestamp to a timezone-aware 
        timestamp in the provided timezone before rendering that timestamp in UTC.
        The function add supports for a majority of timezone abbreviations         

        Arguments:
            timestamp: column or Name
            tz: A String with the time zone ID
        """
        if isinstance(tz, str):
            tz = timezoneMap.get(tz, tz)
            tz = F.lit(tz)
        elif isinstance(tz,Column):
            tz = map_timestamp(tz)
        timestamp = _to_col_if_str(timestamp,"to_utc_timestamp")
        return F.to_utc_timestamp(timestamp,tz) 



    F.array          = _array
    F.array_flatten  = _array_flatten
    F.arrays_zip     = _arrays_zip
    F.create_map     = create_map
    F.format_number  = format_number
    F.flatten        = _array_flatten
    F.map_values     = _map_values
    F.regexp_split   = _regexp_split
    F.to_utc_timestamp_ext = to_utc_timestamp_ext


from snowflake.snowpark import Session

if not hasattr(Session,"___extended"):
    setattr(Session,"___extended",True)
    def __repr__(self):
        from snowflake.snowpark import Session
        from snowflake.snowpark.version import VERSION
        snowflake_environment = self.sql('select current_version()').collect()
        snowpark_version = VERSION
        # Current Environment Details
        return f'Role                        : {self.get_current_role()}\n' +\
        f'Warehouse                   : {self.get_current_warehouse()}\n' +\
        f'Database                    : {self.get_current_database()}\n' +\
        f'Schema                      : {self.get_current_schema()}\n' +\
        f'Snowflake version           : {snowflake_environment[0][0]}\n' +\
        f'Snowpark for Python version : {snowpark_version[0]}.{snowpark_version[1]}.{snowpark_version[2]}\n'
    
    # this method is added for IPython compatibility
    setattr(Session,"__repr__",__repr__)